@@ -1,10 +1,10 @@
-use asdi::idb::{Evaluator, NaiveEvaluator};
+use asdi::idb::eval::{Evaluator, NaiveEvaluator};
 use asdi::io::{print_relation, Format};
 use asdi::parse::parse_file;
 use asdi::{Collection, Labeled};
 
 fn main() {
-    let mut program = parse_file("examples/reasoning.dl").unwrap().into_parsed();
+    let mut program = parse_file("examples/rdfs.dl").unwrap().into_parsed();
 
     program.load_extensional_data().unwrap();
 
