@@ -18,7 +18,8 @@ We can deduce:
 1. This relation has _at least_ one fact which has the [constant](Constant) values
    `("Alice", 49, true)`.
 
-If we were to include an extensional relation declaration in our example, as follows:
+If we were to include an extensional relation declaration, using the `assert` pragma, in our
+example from above:
 
 ```datalog
 .assert person(name: string, age: integer, funny: boolean).
@@ -26,9 +27,21 @@ If we were to include an extensional relation declaration in our example, as fol
 person("Alice", 49, true).
 ```
 
-We can add the [label](Predicate) for each attribute of `person` to the relation's schema.
+we may now add the [label](Predicate) for each attribute of `person` to the relation's schema.
 
-# Example
+Sometimes, where the set of facts for an extensional relation comes from an external source, _or_
+is somewhat large it is useful to place the data in an external file. The `input` pragma allows
+us to reference a single file that contains data for a relation.
+
+```datalog
+.assert person(name: string, age: integer, funny: boolean).
+.input(person, "data/people.csv", "csv").
+```
+
+Note that any use of `input` requires that the relation exists in the extensional database, either
+by means of the `assert` pragma, or by having previously asserted facts for this relation.
+
+# Examples
 
 The following declares two extensional [relations](Relation), `human` and `age`.
 
@@ -37,6 +50,35 @@ The following declares two extensional [relations](Relation), `human` and `age`.
 .assert age(name: string, integer).
 ```
 
+The following is the API equivalent of the Datalog program above.
+
+```rust
+use asdi::edb::{Attribute, Predicate, Relation, RelationSet};
+use std::str::FromStr;
+use asdi::{NameReferenceSet, Program};
+
+let mut program = Program::default();
+
+let human = program.predicates().fetch("human").unwrap();
+program.extensional_mut().add_new_relation(
+    human,
+    vec![Attribute::string()]);
+
+let relation = Relation::new(
+    program
+        .predicates()
+        .fetch("age")
+        .unwrap(),
+    vec![
+        Attribute::string_labeled(program
+            .predicates()
+            .fetch("name")
+            .unwrap()),
+        Attribute::integer()
+    ]);
+program.extensional_mut().add(relation);
+```
+
 The following are all valid [facts](Fact) expressed in the text representation.
 
 ```datalog
@@ -51,18 +93,21 @@ boolean_fact(false).
 */
 
 use crate::error::{
-    fact_does_not_correspond_to_schema, invalid_value, relation_exists, Error, Result,
+    fact_does_not_correspond_to_schema, invalid_value, nullary_facts_not_allowed, relation_exists,
+    Error, Result,
 };
-use crate::idb::{Atom, Row, View};
+use crate::idb::query::relational::{FactOps, Projection, Selection};
+use crate::idb::query::{Queryable, Row, View};
+use crate::idb::Atom;
 use crate::io::{read_relation, write_relation, FilePragma};
 use crate::syntax::{
-    CHAR_COLON, CHAR_COMMA, CHAR_LEFT_PAREN, CHAR_PERIOD, CHAR_RIGHT_PAREN, CHAR_UNDERSCORE,
-    COLUMN_NAME_UNKNOWN, RESERVED_BOOLEAN_FALSE, RESERVED_BOOLEAN_TRUE, RESERVED_PRAGMA_ASSERT,
-    RESERVED_PRAGMA_INFER, RESERVED_PREFIX, TYPE_NAME_PREDICATE,
+    ANONYMOUS_COLUMN_NAME, BOOLEAN_LITERAL_FALSE, BOOLEAN_LITERAL_TRUE, CHAR_COLON, CHAR_COMMA,
+    CHAR_LEFT_PAREN, CHAR_PERIOD, CHAR_RIGHT_PAREN, CHAR_UNDERSCORE, PRAGMA_ID_ASSERT,
+    PRAGMA_ID_INFER, RESERVED_PREFIX, TYPE_NAME_PREDICATE,
 };
 use crate::{
-    AttributeName, Collection, IndexedCollection, Labeled, MaybeAnonymous, MaybeLabeled, Term,
-    Variable,
+    AttributeName, AttributeNameRef, Collection, IndexedCollection, Labeled, MaybeAnonymous,
+    MaybeLabeled, Term, Variable,
 };
 use paste::paste;
 use std::collections::{BTreeMap, HashSet};
@@ -79,7 +124,7 @@ use tracing::trace;
 /// A Set-like collection of [`Relation`]s, the collection is indexed by the relation's label.
 ///
 #[derive(Clone, Debug, Default, PartialEq, Eq)]
-pub struct Relations(BTreeMap<Predicate, Relation>);
+pub struct RelationSet(BTreeMap<Predicate, Relation>);
 
 ///
 /// A Relation comprises a [label](struct.Predicate.html), a [schema](struct.Schema.html) that
@@ -98,25 +143,26 @@ pub struct Relation {
 ///
 /// A schema is an ordered list of [attribute descriptions](Attribute).
 ///
-#[derive(Debug, Clone, PartialEq, Eq)]
+#[derive(Debug, Clone, PartialEq, Eq, Hash)]
 pub struct Schema<T>
 where
     T: AttributeName,
 {
     attributes: Vec<Attribute<T>>,
-    label_index: BTreeMap<T, usize>,
+    label_index: BTreeMap<AttributeNameRef<T>, usize>,
 }
 
 ///
 /// An Attribute structure provides the label and [type](AttributeKind) of an attribute
 /// within a relation.
 ///
-#[derive(Clone, Debug, PartialEq, Eq)]
+#[derive(Clone, Debug, PartialEq, Eq, Hash)]
 pub struct Attribute<T>
 where
     T: AttributeName,
 {
-    label: Option<T>,
+    index: Option<usize>,
+    label: Option<AttributeNameRef<T>>,
     kind: Option<AttributeKind>,
 }
 
@@ -129,14 +175,14 @@ pub enum AttributeIndex<T>
 where
     T: AttributeName,
 {
-    Label(T),
+    Label(AttributeNameRef<T>),
     Index(usize),
 }
 
 ///
 /// The currently supported set of types for [attributes](Attribute).
 ///
-#[derive(Clone, Copy, Debug, PartialEq, Eq)]
+#[derive(Clone, Copy, Debug, PartialEq, Eq, Hash)]
 pub enum AttributeKind {
     String,
     Integer,
@@ -166,6 +212,8 @@ pub enum Constant {
     Boolean(bool),
 }
 
+// ------------------------------------------------------------------------------------------------
+
 ///
 /// A predicate is a value from the set $\small \mathcal{P}$ and labels [relations](struct.Relation.html),
 /// and [atoms](../idb/Atom), and [facts](struct.Fact.html) associated with relations.
@@ -176,6 +224,14 @@ pub enum Constant {
 #[derive(Clone, Debug, PartialEq, Eq, PartialOrd, Ord, Hash)]
 pub struct Predicate(String);
 
+///
+/// We use a reference-counted type for predicates to try and reduce the number of instances of a
+/// commonly used value.
+///
+/// The type [PredicateSet](../struct.PredicateSet.html), and the program's
+/// shared set via [predicates](../struct.PredicateSet.html#method.predicates) allows for common
+/// memory usage within a program.
+///
 pub type PredicateRef = Rc<Predicate>;
 
 // ------------------------------------------------------------------------------------------------
@@ -237,13 +293,9 @@ impl<T> Collection<Attribute<T>> for Schema<T>
 where
     T: AttributeName,
 {
-    fn is_empty(&self) -> bool {
-        self.attributes.is_empty()
-    }
+    delegate!(is_empty, attributes -> bool);
 
-    fn len(&self) -> usize {
-        self.attributes.len()
-    }
+    delegate!(len, attributes -> usize);
 
     fn iter(&self) -> Box<dyn Iterator<Item = &'_ Attribute<T>> + '_> {
         Box::new(self.attributes.iter())
@@ -282,7 +334,7 @@ where
 {
     pub fn new<V: Into<Vec<Attribute<T>>>>(attributes: V) -> Self {
         let attributes = attributes.into();
-        let named: Vec<(T, usize)> = attributes
+        let named: Vec<(AttributeNameRef<T>, usize)> = attributes
             .iter()
             .enumerate()
             .filter_map(|(i, c)| c.label().map(|var| (var.clone(), i)))
@@ -316,11 +368,11 @@ where
 
     // --------------------------------------------------------------------------------------------
 
-    pub fn labels(&self) -> impl Iterator<Item = &T> {
+    pub fn labels(&self) -> impl Iterator<Item = &AttributeNameRef<T>> {
         self.label_index.keys()
     }
 
-    pub fn label_intersection(&self, other: &Self) -> Vec<(&T, usize, usize)> {
+    pub fn label_intersection(&self, other: &Self) -> Vec<(&AttributeNameRef<T>, usize, usize)> {
         self.label_index
             .iter()
             .filter_map(|(var, i)| {
@@ -332,8 +384,8 @@ where
             .collect()
     }
 
-    pub fn label_union(&self, other: &Self) -> Vec<T> {
-        let mut all: Vec<T> = self.labels().cloned().collect();
+    pub fn label_union(&self, other: &Self) -> Vec<AttributeNameRef<T>> {
+        let mut all: Vec<AttributeNameRef<T>> = self.labels().cloned().collect();
         for var in other.labels() {
             if !all.contains(var) {
                 all.push(var.clone());
@@ -378,12 +430,12 @@ where
     }
 }
 
-impl<T> From<T> for Attribute<T>
+impl<T> From<AttributeNameRef<T>> for Attribute<T>
 where
     T: AttributeName,
 {
-    fn from(name: T) -> Self {
-        Self::labeled(name)
+    fn from(label: AttributeNameRef<T>) -> Self {
+        Self::labeled(label)
     }
 }
 
@@ -403,7 +455,7 @@ impl<T> MaybeLabeled<T> for Attribute<T>
 where
     T: AttributeName,
 {
-    fn label(&self) -> Option<&T> {
+    fn label(&self) -> Option<&AttributeNameRef<T>> {
         self.label.as_ref()
     }
 }
@@ -412,7 +464,7 @@ impl<T> Attribute<T>
 where
     T: AttributeName,
 {
-    pub fn labeled(name: T) -> Self {
+    pub fn labeled(name: AttributeNameRef<T>) -> Self {
         Self::new_inner(name, None)
     }
 
@@ -424,7 +476,7 @@ where
         Self::new_inner(None, AttributeKind::String)
     }
 
-    pub fn string_labeled(name: T) -> Self {
+    pub fn string_labeled(name: AttributeNameRef<T>) -> Self {
         Self::new_inner(name, AttributeKind::String)
     }
 
@@ -432,7 +484,7 @@ where
         Self::new_inner(None, AttributeKind::Integer)
     }
 
-    pub fn integer_labeled(name: T) -> Self {
+    pub fn integer_labeled(name: AttributeNameRef<T>) -> Self {
         Self::new_inner(name, AttributeKind::Integer)
     }
 
@@ -440,24 +492,36 @@ where
         Self::new_inner(None, AttributeKind::Boolean)
     }
 
-    pub fn boolean_labeled(name: T) -> Self {
+    pub fn boolean_labeled(name: AttributeNameRef<T>) -> Self {
         Self::new_inner(name, AttributeKind::Boolean)
     }
 
-    pub fn new(name: T, kind: AttributeKind) -> Self {
+    pub fn new(name: AttributeNameRef<T>, kind: AttributeKind) -> Self {
         Self::new_inner(name, kind)
     }
 
-    pub(crate) fn new_inner<P: Into<Option<T>>, K: Into<Option<AttributeKind>>>(
+    pub(crate) fn new_inner<
+        P: Into<Option<AttributeNameRef<T>>>,
+        K: Into<Option<AttributeKind>>,
+    >(
         name: P,
         kind: K,
     ) -> Self {
         Self {
+            index: None,
             label: name.into(),
             kind: kind.into(),
         }
     }
 
+    pub fn index(&self) -> Option<usize> {
+        self.index
+    }
+
+    pub fn set_index(&mut self, index: usize) {
+        self.index = Some(index)
+    }
+
     pub fn kind(&self) -> Option<AttributeKind> {
         self.kind.as_ref().copied()
     }
@@ -488,7 +552,7 @@ where
             (Some(n), None) => format!("{}", n),
             (None, Some(k)) => {
                 if emit_unknown {
-                    format!("{}{} {}", COLUMN_NAME_UNKNOWN, CHAR_COLON, k)
+                    format!("{}{} {}", ANONYMOUS_COLUMN_NAME, CHAR_COLON, k)
                 } else {
                     k.to_string()
                 }
@@ -508,11 +572,11 @@ where
     }
 }
 
-impl<T> From<T> for AttributeIndex<T>
+impl<T> From<AttributeNameRef<T>> for AttributeIndex<T>
 where
     T: AttributeName,
 {
-    fn from(v: T) -> Self {
+    fn from(v: AttributeNameRef<T>) -> Self {
         Self::Label(v)
     }
 }
@@ -561,7 +625,9 @@ impl From<Constant> for AttributeKind {
     }
 }
 
-impl<'a> IntoIterator for &'a mut Relations {
+// ------------------------------------------------------------------------------------------------
+
+impl<'a> IntoIterator for &'a mut RelationSet {
     type Item = (&'a Predicate, &'a mut Relation);
     type IntoIter = std::collections::btree_map::IterMut<'a, Predicate, Relation>;
 
@@ -570,7 +636,7 @@ impl<'a> IntoIterator for &'a mut Relations {
     }
 }
 
-impl FromIterator<Relation> for Relations {
+impl FromIterator<Relation> for RelationSet {
     fn from_iter<T: IntoIterator<Item = Relation>>(iter: T) -> Self {
         Self(BTreeMap::from_iter(
             iter.into_iter().map(|r| (r.label().clone(), r)),
@@ -578,14 +644,20 @@ impl FromIterator<Relation> for Relations {
     }
 }
 
-impl Collection<Relation> for Relations {
-    fn is_empty(&self) -> bool {
-        self.0.is_empty()
+impl Queryable for RelationSet {
+    fn query_atom(&self, query: &Atom) -> Result<Option<View>> {
+        if let Some(relation) = self.0.get(query.label()) {
+            relation.query_atom(query)
+        } else {
+            Ok(None)
+        }
     }
+}
 
-    fn len(&self) -> usize {
-        self.0.len()
-    }
+impl Collection<Relation> for RelationSet {
+    delegate!(is_empty -> bool);
+
+    delegate!(len -> usize);
 
     fn iter(&self) -> Box<dyn Iterator<Item = &'_ Relation> + '_> {
         Box::new(self.0.values())
@@ -596,7 +668,7 @@ impl Collection<Relation> for Relations {
     }
 }
 
-impl IndexedCollection<Predicate, Relation> for Relations {
+impl IndexedCollection<Predicate, Relation> for RelationSet {
     fn get(&self, index: &Predicate) -> Option<&Relation> {
         self.0.get(index)
     }
@@ -606,7 +678,7 @@ impl IndexedCollection<Predicate, Relation> for Relations {
     }
 }
 
-impl Relations {
+impl RelationSet {
     pub fn flat_count(&self) -> usize {
         self.iter().map(|r| r.len()).sum()
     }
@@ -675,29 +747,6 @@ impl Relations {
         self.0.get_mut(predicate)
     }
 
-    pub fn matches(&self, atom: &Atom) -> Option<View> {
-        if let Some(relation) = self.0.get(atom.label()) {
-            trace!(
-                "matches > predicate: {}, relation: {:?}",
-                atom.label(),
-                relation.label()
-            );
-            let results = relation.matches(atom);
-            if atom.variables().count() == 0 {
-                // if all attributes are constant it is a presence query, not a selection.
-                if results.is_empty() {
-                    Some(View::new_false())
-                } else {
-                    Some(View::new_true())
-                }
-            } else {
-                Some(results)
-            }
-        } else {
-            None
-        }
-    }
-
     pub fn clone_with_schema_only(&self) -> Self {
         Self(
             self.0
@@ -707,38 +756,40 @@ impl Relations {
         )
     }
 
-    pub fn merge(&mut self, other: Self) -> Result<()> {
+    pub fn merge_from(&mut self, other: Self) -> Result<()> {
         for (other_predicate, other_relation) in other.0 {
             if let Some(relation) = self.0.get_mut(&other_predicate) {
-                relation.extend(other_relation)?
+                relation.extend(other_relation)?;
             } else {
                 self.0.insert(other_predicate, other_relation);
             }
         }
         Ok(())
     }
+
+    pub fn merge(&mut self, other: &Self) -> Result<()> {
+        for (other_predicate, other_relation) in &other.0 {
+            if let Some(relation) = self.0.get_mut(other_predicate) {
+                for fact in other_relation.facts.iter() {
+                    relation.add(fact.clone())?;
+                }
+            } else {
+                self.0
+                    .insert(other_predicate.clone(), other_relation.clone());
+            }
+        }
+        Ok(())
+    }
 }
 
 // ------------------------------------------------------------------------------------------------
 
-impl Labeled for Relation {
-    fn label(&self) -> &Predicate {
-        &self.label
-    }
-
-    fn label_ref(&self) -> PredicateRef {
-        self.label.clone()
-    }
-}
+impl_labeled! {Relation}
 
 impl Collection<Fact> for Relation {
-    fn is_empty(&self) -> bool {
-        self.facts.is_empty()
-    }
+    delegate!(is_empty, facts -> bool);
 
-    fn len(&self) -> usize {
-        self.facts.len()
-    }
+    delegate!(len, facts -> usize);
 
     fn iter(&self) -> Box<dyn Iterator<Item = &'_ Fact> + '_> {
         Box::new(self.facts.iter())
@@ -749,6 +800,49 @@ impl Collection<Fact> for Relation {
     }
 }
 
+impl Queryable for Relation {
+    fn query_atom(&self, query: &Atom) -> Result<Option<View>> {
+        let schema = self.make_view_schema(&query.iter().collect::<Vec<&Term>>());
+
+        let view = if query.is_universal() {
+            // all terms are variables, so all facts match
+            View::new_with_facts(
+                schema,
+                self.facts
+                    .iter()
+                    .map(|fact| Row::from(fact.values().clone()))
+                    .collect::<Vec<Row>>(),
+            )
+        } else if query.is_existential() {
+            // all terms are constants, so only one term should match
+            let fact = Fact::new(
+                query.label_ref(),
+                query.constants().cloned().collect::<Vec<Constant>>(),
+            )?;
+            if self.contains(&fact) {
+                View::new_true()
+            } else {
+                View::new_false()
+            }
+        } else {
+            let selection = Selection::try_from(query)?;
+            // let projection = Projection::try_from(query)?;
+            let rows: Result<Vec<Row>> = self
+                .iter()
+                .filter_map(|fact| fact.clone().select(&selection).transpose())
+                // TODO: Need to update the view join logic
+                // .map(|fact| match fact {
+                //     Ok(fact) => fact.project(&projection),
+                //     Err(e) => Err(e),
+                // })
+                .collect();
+            View::new_with_facts(schema, rows?)
+        };
+
+        Ok(Some(view))
+    }
+}
+
 impl Relation {
     pub fn new<V: Into<Schema<Predicate>>>(name: PredicateRef, schema: V) -> Self {
         Self {
@@ -795,11 +889,10 @@ impl Relation {
 
     // --------------------------------------------------------------------------------------------
 
-    pub fn add_as_fact<V: Into<Vec<Constant>>>(&mut self, values: V) -> Result<()> {
+    pub fn add_as_fact<V: Into<Vec<Constant>>>(&mut self, values: V) -> Result<bool> {
         let values: Vec<Constant> = values.into();
         if self.schema.conforms(&values) {
-            self.add(Fact::new(self.label.clone(), values))?;
-            Ok(())
+            Ok(self.add(Fact::new(self.label.clone(), values)?)?)
         } else {
             Err(fact_does_not_correspond_to_schema(
                 self.label_ref(),
@@ -812,13 +905,9 @@ impl Relation {
         }
     }
 
-    pub fn add(&mut self, fact: Fact) -> Result<()> {
+    pub fn add(&mut self, fact: Fact) -> Result<bool> {
         if fact.label() == self.label() {
-            let fact_str = fact.to_string();
-            if !self.facts.insert(fact) {
-                trace!("Fact {} discarded, already present in relation.", fact_str);
-            }
-            Ok(())
+            Ok(self.facts.insert(fact))
         } else {
             Err(fact_does_not_correspond_to_schema(
                 fact.label_ref(),
@@ -830,12 +919,13 @@ impl Relation {
         }
     }
 
-    pub fn extend(&mut self, other: Self) -> Result<()> {
+    pub fn extend(&mut self, other: Self) -> Result<usize> {
         if self.label() == other.label() && self.schema() == other.schema() {
+            let mut added = 0;
             for fact in other.facts.into_iter() {
-                self.add(fact)?;
+                added += if self.add(fact)? { 1 } else { 0 };
             }
-            Ok(())
+            Ok(added)
         } else {
             panic!()
         }
@@ -843,53 +933,25 @@ impl Relation {
 
     // --------------------------------------------------------------------------------------------
 
-    pub(crate) fn matches(&self, atom: &Atom) -> View {
-        let terms: Vec<&Term> = atom.iter().collect();
-        self.match_terms(terms)
-    }
-
-    #[allow(single_use_lifetimes)]
-    fn match_terms<'a, V: Into<Vec<&'a Term>>>(&self, terms: V) -> View {
-        let terms = terms.into();
-
-        View::new_with_facts(
-            terms
-                .iter()
-                .enumerate()
-                .map(|(i, term)| {
-                    let mut term = match term {
-                        Term::Anonymous => Attribute::anonymous(),
-                        Term::Variable(v) => Attribute::from(v.clone()),
-                        Term::Constant(v) => Attribute::typed(v.kind()),
-                    };
-                    if term.kind().is_none() {
-                        if let Some(kind) =
-                            self.schema().get(&AttributeIndex::Index(i)).unwrap().kind()
-                        {
-                            term.override_kind(kind);
-                        }
-                    }
-                    term
-                })
-                .collect::<Vec<Attribute<Variable>>>(),
-            self.facts
-                .iter()
-                .filter_map(|fact| self.match_terms_inner(&terms, fact.values()))
-                .collect::<Vec<Row>>(),
-        )
-    }
-
-    fn match_terms_inner(&self, terms: &[&Term], fact: &[Constant]) -> Option<Row> {
-        if terms
+    fn make_view_schema(&self, terms: &[&Term]) -> Vec<Attribute<Variable>> {
+        terms
             .iter()
             .enumerate()
-            .filter(|(_, term)| term.is_constant())
-            .all(|(i, term)| term.as_constant().unwrap() == fact.get(i).unwrap())
-        {
-            Some(Row::from(fact.to_vec()))
-        } else {
-            None
-        }
+            .map(|(i, term)| {
+                let mut term = match term {
+                    Term::Anonymous => Attribute::anonymous(),
+                    Term::Variable(v) => Attribute::from(v.clone()),
+                    Term::Constant(v) => Attribute::typed(v.kind()),
+                };
+                if term.kind().is_none() {
+                    if let Some(kind) = self.schema().get(&AttributeIndex::Index(i)).unwrap().kind()
+                    {
+                        term.override_kind(kind);
+                    }
+                }
+                term
+            })
+            .collect()
     }
 
     // --------------------------------------------------------------------------------------------
@@ -919,17 +981,17 @@ impl Relation {
     /// If a [FilePragma] is attached to this relation then read data from the file into this
     /// relation's set of facts.
     ///
-    pub fn load_from_file(&mut self) -> Result<()> {
+    pub fn load_from_file(&mut self) -> Result<usize> {
         if let Some(pragma) = &self.pragma {
             let new_facts = read_relation(self, pragma)?;
             self.extend(new_facts)
         } else {
-            Ok(())
+            Ok(0)
         }
     }
 
     ///
-    /// If a file pragma is attached to this relation then store this relation's set of facts
+    /// If a [FilePragma] is attached to this relation then store this relation's set of facts
     /// into the file.
     ///
     pub fn store_to_file(&mut self) -> Result<()> {
@@ -947,9 +1009,9 @@ impl Relation {
             "{}{} {}{}{}{}{}",
             RESERVED_PREFIX,
             if extensional {
-                RESERVED_PRAGMA_ASSERT
+                PRAGMA_ID_ASSERT
             } else {
-                RESERVED_PRAGMA_INFER
+                PRAGMA_ID_INFER
             },
             self.label(),
             CHAR_LEFT_PAREN,
@@ -985,56 +1047,57 @@ impl From<Fact> for Vec<Constant> {
     }
 }
 
-impl Labeled for Fact {
-    fn label(&self) -> &Predicate {
-        &self.label
+impl_labeled! {Fact}
+
+impl_collection! {Fact, Constant, values}
+
+impl_indexed_collection! { Fact, Constant, usize, values}
+
+impl FactOps for Fact {
+    fn project(self, projection: &Projection) -> Result<Row> {
+        Ok(if projection.is_all() {
+            Row::from(self)
+        } else {
+            Row::from(
+                self.values
+                    .into_iter()
+                    .enumerate()
+                    .filter_map(|(i, c)| {
+                        if projection.contains_index(&i) {
+                            Some(c)
+                        } else {
+                            None
+                        }
+                    })
+                    .collect::<Vec<Constant>>(),
+            )
+        })
     }
 
-    fn label_ref(&self) -> PredicateRef {
-        self.label.clone()
-    }
-}
-
-impl Collection<Constant> for Fact {
-    fn is_empty(&self) -> bool {
-        self.values.is_empty()
-    }
-
-    fn len(&self) -> usize {
-        self.values.len()
-    }
-
-    fn iter(&self) -> Box<dyn Iterator<Item = &'_ Constant> + '_> {
-        Box::new(self.values.iter())
-    }
-
-    fn contains(&self, value: &Constant) -> bool {
-        self.values.contains(value)
-    }
-}
-
-impl IndexedCollection<usize, Constant> for Fact {
-    fn get(&self, index: &usize) -> Option<&Constant> {
-        self.values.get(*index)
-    }
-
-    fn contains_index(&self, index: &usize) -> bool {
-        *index < self.len()
+    fn select(self, criteria: &Selection) -> Result<Option<Row>> {
+        Ok(if criteria.is_all() || criteria.matches(self.values())? {
+            Some(Row::from(self))
+        } else {
+            None
+        })
     }
 }
 
 impl Fact {
-    pub fn new<V: Into<Vec<Constant>>>(name: PredicateRef, values: V) -> Self {
-        Self {
-            label: name,
-            values: values.into(),
+    pub fn new<V: Into<Vec<Constant>>>(label: PredicateRef, values: V) -> Result<Self> {
+        let values = values.into();
+        if values.is_empty() {
+            Err(nullary_facts_not_allowed())
+        } else {
+            Ok(Self { label, values })
         }
     }
 
-    fn values(&self) -> &Vec<Constant> {
+    pub fn values(&self) -> &Vec<Constant> {
         &self.values
     }
 }
+
 // ------------------------------------------------------------------------------------------------
 
 impl From<&str> for Constant {
@@ -1053,17 +1116,9 @@ impl From<String> for Constant {
     }
 }
 
-impl From<i64> for Constant {
-    fn from(v: i64) -> Self {
-        Self::Integer(v)
-    }
-}
+impl_enum_from!(Constant, i64, Integer);
 
-impl From<bool> for Constant {
-    fn from(v: bool) -> Self {
-        Self::Boolean(v)
-    }
-}
+impl_enum_from!(Constant, bool, Boolean);
 
 impl Display for Constant {
     fn fmt(&self, f: &mut Formatter<'_>) -> std::fmt::Result {
@@ -1080,10 +1135,11 @@ impl Display for Constant {
                 Self::Integer(v) => v.to_string(),
                 Self::Boolean(v) => {
                     if *v {
-                        format!("{}{}", RESERVED_PREFIX, RESERVED_BOOLEAN_TRUE)
+                        BOOLEAN_LITERAL_TRUE
                     } else {
-                        format!("{}{}", RESERVED_PREFIX, RESERVED_BOOLEAN_FALSE)
+                        BOOLEAN_LITERAL_FALSE
                     }
+                    .to_string()
                 }
             }
         )
@@ -1103,6 +1159,12 @@ impl Constant {
         Self::Integer(v as i64)
     }
 
+    self_is_as!(string, String);
+
+    self_is_as!(integer, Integer, i64);
+
+    self_is_as!(boolean, Boolean, bool);
+
     pub fn kind(&self) -> AttributeKind {
         match self {
             Self::String(_) => AttributeKind::String,
@@ -1157,12 +1219,6 @@ impl AsRef<str> for Predicate {
     }
 }
 
-impl From<Predicate> for String {
-    fn from(v: Predicate) -> Self {
-        v.0
-    }
-}
-
 impl AttributeName for Predicate {
     fn is_valid(s: &str) -> bool {
         let mut chars = s.chars();
@@ -1170,10 +1226,12 @@ impl AttributeName for Predicate {
             && chars.next().map(|c| c.is_lowercase()).unwrap()
             && chars.all(|c| c.is_alphanumeric() || c == CHAR_UNDERSCORE)
     }
+
+    fn type_name() -> &'static str {
+        TYPE_NAME_PREDICATE
+    }
 }
 
 impl Predicate {
-    pub(crate) fn from_str_unchecked(s: &str) -> Self {
-        Self(s.to_owned())
-    }
+    into_inner_fn!(String);
 }
