@@ -102,6 +102,33 @@ pub enum Error {
     RelationDoesNotExist {
         label: PredicateRef,
     },
+
+    AttributeDoesNotExist {
+        label: String,
+    },
+
+    AttributeIndexInvalid {
+        index: usize,
+    },
+
+    NotStratifiable,
+
+    NullaryFactsNotAllowed,
+
+    ComparisonIsAlwaysTrue {
+        comparison: String,
+    },
+
+    ComparisonIsAlwaysFalse {
+        comparison: String,
+    },
+
+    IncompatibleTypes {
+        lhs_type: String,
+        rhs_type: String,
+    },
+
+    AnonymousVariableNotAllowed,
 }
 
 ///
@@ -212,6 +239,54 @@ pub fn relation_does_not_exist(label: PredicateRef) -> Error {
     Error::RelationDoesNotExist { label }
 }
 
+#[inline]
+pub fn attribute_does_not_exist<S: Into<String>>(label: S) -> Error {
+    Error::AttributeDoesNotExist {
+        label: label.into(),
+    }
+}
+
+#[inline]
+pub fn attribute_index_invalid(index: usize) -> Error {
+    Error::AttributeIndexInvalid { index }
+}
+
+#[inline]
+pub fn program_not_stratifiable() -> Error {
+    Error::NotStratifiable
+}
+
+#[inline]
+pub fn nullary_facts_not_allowed() -> Error {
+    Error::NullaryFactsNotAllowed
+}
+
+#[inline]
+pub fn comparison_is_always_true<S: Into<String>>(comparison: S) -> Error {
+    Error::ComparisonIsAlwaysTrue {
+        comparison: comparison.into(),
+    }
+}
+
+#[inline]
+pub fn comparison_is_always_false<S: Into<String>>(comparison: S) -> Error {
+    Error::ComparisonIsAlwaysFalse {
+        comparison: comparison.into(),
+    }
+}
+
+#[inline]
+pub fn anonymous_variable_not_allowed() -> Error {
+    Error::AnonymousVariableNotAllowed
+}
+
+pub fn incompatible_types<S: Into<String>>(lhs_type: S, rhs_type: S) -> Error {
+    Error::IncompatibleTypes {
+        lhs_type: lhs_type.into(),
+        rhs_type: rhs_type.into(),
+    }
+}
+
 // ------------------------------------------------------------------------------------------------
 // Implementations
 // ------------------------------------------------------------------------------------------------
@@ -301,6 +376,14 @@ impl Display for Error {
                 Error::Serialization(e) => format!("An error occured either serializing or deserializing a relation: {}", e),
                 Error::SerializationFormatUnknown { format: serialization } => format!("'{}' is not a supported serialization format.", serialization),
                 Error::SerializationOperationUnsupported{ format: serialization } => format!("The requested I/O operation is not supported by the serialization format '{}'", serialization),
+                Error::NotStratifiable => format!("The program cannot be evaluated as it includes negation but cannot be stratified."),
+                Error::AttributeDoesNotExist { label } => format!("The attribute labeled '{}' was not a member of the target schema.", label),
+                Error::AttributeIndexInvalid { index } => format!("The attribute index '{}' is not valid for the target schema.", index),
+                Error::NullaryFactsNotAllowed => format!("The arity of facts must be greater than, or equal to, 1."),
+                Error::ComparisonIsAlwaysTrue { comparison } => format!("A comparison operator, or selection criteria, will always be true/⊤ (`{}`).", comparison),
+                Error::ComparisonIsAlwaysFalse { comparison } => format!("A comparison operator, or selection criteria, will always be true/⊥ (`{}`).", comparison),
+                Error::AnonymousVariableNotAllowed => format!("Anonymous variables not allowed in the current context."),
+                Error::IncompatibleTypes { lhs_type, rhs_type } => format!("A required operation cannot be performed between values of type `{}` and `{}`.", lhs_type, rhs_type),
             }
         )
     }
