@@ -145,7 +145,7 @@ As defined by Codd, selection is written as $\small \sigma_{a\theta b}(R)$ or
 $\small \sigma_{a\theta v}(R)$ where:
 
 * $\small a$ and $\small b$ are attribute names,
-* $\small\theta$ is a binary operation in the set $\small\lbrace =, \neq, <, \leq, >, \geq \rbrace$,
+* $\small\theta$ is a binary operation, where $\small\theta\in\lbrace =, \neq, <, \leq, >, \geq, \overset{\star}{=} \rbrace$,
 * $\small v$ is a constant value,
 * $\small R$ is a relation,
 
@@ -270,26 +270,32 @@ converted to Datalog using the examples above.
 
 use crate::edb::{AttributeIndex, Constant, Predicate, Relation};
 use crate::error::{
-    fact_does_not_correspond_to_schema, head_variables_missing_in_body, invalid_head_atom_count,
-    invalid_value, negative_variables_not_also_positive, Error, Result, SourceLocation,
+    anonymous_variable_not_allowed, comparison_is_always_false, comparison_is_always_true,
+    fact_does_not_correspond_to_schema, head_variables_missing_in_body, incompatible_types,
+    invalid_head_atom_count, invalid_value, negative_variables_not_also_positive, Error, Result,
+    SourceLocation,
 };
 use crate::features::{FEATURE_CONSTRAINTS, FEATURE_DISJUNCTION};
 use crate::syntax::{
     ANONYMOUS_TERM, CHAR_LEFT_PAREN, CHAR_PERIOD, CHAR_RIGHT_PAREN, CHAR_UNDERSCORE,
     CONJUNCTION_UNICODE_SYMBOL, DISJUNCTION_UNICODE_SYMBOL, EMPTY_STR, FALSE_UNICODE_SYMBOL,
-    IMPLICATION_UNICODE_ARROW, NOT_UNICODE_SYMBOL, OPERATOR_ASCII_EQUAL,
-    OPERATOR_ASCII_GREATER_THAN, OPERATOR_ASCII_GREATER_THAN_OR_EQUAL, OPERATOR_ASCII_LESS_THAN,
-    OPERATOR_ASCII_LESS_THAN_OR_EQUAL, OPERATOR_ASCII_NOT_EQUAL, OPERATOR_ASCII_NOT_EQUAL_ALT,
-    OPERATOR_UNICODE_GREATER_THAN_OR_EQUAL, OPERATOR_UNICODE_LESS_THAN_OR_EQUAL,
-    OPERATOR_UNICODE_NOT_EQUAL, TYPE_NAME_COMPARISON_OPERATOR, TYPE_NAME_VARIABLE,
+    IMPLICATION_UNICODE_ARROW, NEGATION_UNICODE_SYMBOL, OPERATOR_EQUAL_ASCII,
+    OPERATOR_GREATER_THAN_ASCII, OPERATOR_GREATER_THAN_OR_EQUAL_ASCII,
+    OPERATOR_GREATER_THAN_OR_EQUAL_UNICODE, OPERATOR_LESS_THAN_ASCII,
+    OPERATOR_LESS_THAN_OR_EQUAL_ASCII, OPERATOR_LESS_THAN_OR_EQUAL_UNICODE,
+    OPERATOR_NOT_EQUAL_ASCII, OPERATOR_NOT_EQUAL_ASCII_ALT, OPERATOR_NOT_EQUAL_UNICODE,
+    OPERATOR_STRING_MATCH_ASCII, OPERATOR_STRING_MATCH_ASCII_WORD, OPERATOR_STRING_MATCH_UNICODE,
+    TYPE_NAME_COMPARISON_OPERATOR, TYPE_NAME_VARIABLE,
 };
 use crate::{
     AttributeName, Collection, FeatureSet, IndexedCollection, Labeled, MaybeAnonymous, MaybeGround,
-    MaybePositive, PredicateRef, SyntaxFragments,
+    MaybePositive, PredicateRef,
 };
 use paste::paste;
 use std::collections::HashSet;
 use std::fmt::{Display, Formatter};
+use std::hash::Hash;
+use std::rc::Rc;
 use std::str::FromStr;
 
 // ------------------------------------------------------------------------------------------------
@@ -301,7 +307,7 @@ use std::str::FromStr;
 /// of [Relations](../edb/struct.Relations.html).
 ///
 #[derive(Clone, Debug, Default, PartialEq, Eq)]
-pub struct Rules(HashSet<Rule>);
+pub struct RuleSet(HashSet<Rule>);
 
 ///
 /// An individual rule consists of a set of head [atoms](Atom) and a set of body [literals](Literal).
@@ -319,10 +325,30 @@ pub struct Rule {
     body: Vec<Literal>,
 }
 
+///
+/// This enumeration represents the three forms a rule can take. This is defined by the predicate
+/// $\small form(r)$ in the description of [rules](../index.html#rules), formula (vi).
+///
 #[derive(Clone, Copy, Debug, PartialEq, Eq, Hash)]
 pub enum RuleForm {
+    ///
+    /// A pure rule is one where there is only a single atom in the head, regardless of the body and
+    /// is the only form allowed in the strict language $\small\text{Datalog}$.
+    /// $$\small |head\(r\)| = 1$$
+    ///
     Pure,
+    ///
+    /// A constraint rule, or contradiction, does not allow any consequence to be determined
+    /// from evaluation of its body, it is only available in the language $\small\text{Datalog}^{\lnot}$.
+    /// $$\small |head\(r\)| = 0 \land \text{Datalog}^{\lnot}$$
+    ///
     Constraint,
+    ///
+    /// A disjunctive rule is one where there is more than one atom, and any one (inclusive
+    /// disjunction) may be true if the body is true, it is only available in the language
+    /// $\small\text{Datalog}^{\lor}$.
+    /// $$\small |head\(r\)| > 1  \land \text{Datalog}^{\lor}$$
+    ///
     Disjunctive,
 }
 
@@ -399,9 +425,9 @@ pub struct Literal {
 #[derive(Clone, Debug, PartialEq, Eq, Hash)]
 pub enum LiteralInner {
     /// A relational literal
-    Atom(Atom),
+    Relational(Atom),
     /// An arithmetic literal, if using the language $\small\text{Datalog}^{\theta}$.
-    Comparison(Comparison),
+    Arithmetic(Comparison),
 }
 
 // ------------------------------------------------------------------------------------------------
@@ -419,7 +445,7 @@ pub struct Comparison {
 ///
 /// The supported set of operations within an arithmetic literal.
 ///
-#[derive(Clone, Debug, PartialEq, Eq, Hash)]
+#[derive(Clone, Copy, Debug, PartialEq, Eq, Hash)]
 pub enum ComparisonOperator {
     Equal,
     NotEqual,
@@ -427,6 +453,7 @@ pub enum ComparisonOperator {
     LessThanOrEqual,
     GreaterThan,
     GreaterThanOrEqual,
+    StringMatch,
 }
 
 // ------------------------------------------------------------------------------------------------
@@ -439,7 +466,7 @@ pub enum Term {
     /// denoted with the character `_` in a literal.
     Anonymous,
     /// a value such that term is $\small \in \mathcal{V}$
-    Variable(Variable),
+    Variable(VariableRef),
 
     /// a value such that term is $\small \in \mathcal{C}$
     Constant(Constant),
@@ -454,6 +481,16 @@ pub enum Term {
 #[derive(Clone, Debug, PartialEq, Eq, PartialOrd, Ord, Hash)]
 pub struct Variable(String);
 
+///
+/// We use a reference-counted type for predicates to try and reduce the number of instances of a
+/// commonly used value.
+///
+// The type [PredicateSet](../struct.PredicateSet.html), and the program's
+// shared set via [predicates](../struct.PredicateSet.html#method.predicates) allows for common
+// memory usage within a program.
+//
+pub type VariableRef = Rc<Variable>;
+
 // ------------------------------------------------------------------------------------------------
 // Private Types & Constants
 // ------------------------------------------------------------------------------------------------
@@ -470,7 +507,7 @@ pub struct Variable(String);
 // Implementations
 // ------------------------------------------------------------------------------------------------
 
-impl Display for Rules {
+impl Display for RuleSet {
     fn fmt(&self, f: &mut Formatter<'_>) -> std::fmt::Result {
         for rule in self.iter() {
             writeln!(f, "{}", rule)?;
@@ -479,14 +516,10 @@ impl Display for Rules {
     }
 }
 
-impl Collection<Rule> for Rules {
-    fn is_empty(&self) -> bool {
-        self.0.is_empty()
-    }
+impl Collection<Rule> for RuleSet {
+    delegate!(is_empty -> bool);
 
-    fn len(&self) -> usize {
-        self.0.len()
-    }
+    delegate!(len -> usize);
 
     fn iter(&self) -> Box<dyn Iterator<Item = &'_ Rule> + '_> {
         Box::new(self.0.iter())
@@ -497,7 +530,7 @@ impl Collection<Rule> for Rules {
     }
 }
 
-impl Rules {
+impl RuleSet {
     pub fn add(&mut self, rule: Rule) {
         self.0.insert(rule);
     }
@@ -505,50 +538,6 @@ impl Rules {
 
 // ------------------------------------------------------------------------------------------------
 
-impl SyntaxFragments for Rule {
-    fn is_linear(&self) -> bool {
-        self.body.len() == 1
-    }
-
-    fn is_guarded(&self) -> bool {
-        let all_variables = self.variables();
-        self.literals().any(|lit| {
-            let lit_variables: HashSet<&Variable> = HashSet::from_iter(lit.variables().into_iter());
-            lit_variables == all_variables
-        })
-    }
-
-    fn is_frontier_guarded(&self) -> bool {
-        let frontier_variables: HashSet<&Variable> = self
-            .head_variables()
-            .intersection(&self.variables())
-            .copied()
-            .collect();
-        self.literals().any(|lit| {
-            let lit_variables: HashSet<&Variable> = HashSet::from_iter(lit.variables().into_iter());
-            lit_variables == frontier_variables
-        })
-    }
-
-    fn is_non_recursive(&self) -> bool {
-        // TODO: this is only direct recursion, need to check for mutual recursive rules.
-        let head_predicates = self
-            .head()
-            .map(|atom| atom.label())
-            .collect::<Vec<&Predicate>>();
-        !self
-            .literals()
-            .filter_map(|lit| {
-                if let LiteralInner::Atom(atom) = lit.as_ref() {
-                    Some(atom.label())
-                } else {
-                    None
-                }
-            })
-            .any(|predicate| head_predicates.contains(&predicate))
-    }
-}
-
 impl Display for Rule {
     fn fmt(&self, f: &mut Formatter<'_>) -> std::fmt::Result {
         write!(
@@ -669,6 +658,11 @@ impl Rule {
         self.head().map(|atom| atom.iter()).flatten().collect()
     }
 
+    pub fn distinguished_terms_in_order(&self) -> Vec<&Term> {
+        let terms: Vec<&Term> = self.head().map(|atom| atom.iter()).flatten().collect();
+        dedup_in_place(terms)
+    }
+
     pub fn non_distinguished_terms(&self) -> HashSet<&Term> {
         let distinguished = self.distinguished_terms();
         self.terms()
@@ -683,7 +677,7 @@ impl Rule {
             .map(|lit| lit.terms())
             .flatten()
             .collect::<HashSet<&Term>>()
-            .intersection(&self.distinguished_terms())
+            .union(&self.distinguished_terms())
             .copied()
             .collect()
     }
@@ -708,11 +702,11 @@ impl Rule {
 
     // --------------------------------------------------------------------------------------------
 
-    pub fn head_variables(&self) -> HashSet<&Variable> {
+    pub fn head_variables(&self) -> HashSet<&VariableRef> {
         self.head().map(|atom| atom.variables()).flatten().collect()
     }
 
-    pub fn variables(&self) -> HashSet<&Variable> {
+    pub fn variables(&self) -> HashSet<&VariableRef> {
         self.body
             .iter()
             .map(|lit| lit.variables())
@@ -720,7 +714,7 @@ impl Rule {
             .collect()
     }
 
-    pub fn positive_variables(&self) -> HashSet<&Variable> {
+    pub fn positive_variables(&self) -> HashSet<&VariableRef> {
         self.body
             .iter()
             .filter(|lit| lit.is_positive())
@@ -729,7 +723,7 @@ impl Rule {
             .collect()
     }
 
-    pub fn negative_variables(&self) -> HashSet<&Variable> {
+    pub fn negative_variables(&self) -> HashSet<&VariableRef> {
         self.body
             .iter()
             .filter(|lit| !lit.is_positive())
@@ -741,7 +735,40 @@ impl Rule {
     // --------------------------------------------------------------------------------------------
 
     ///
-    /// # Well-formedness
+    /// Guarded $\small\text{Datalog}$ is defined where for every rule, all the variables that occur
+    /// in the rule bodies must occur together in at least one atom, called a guard atom.
+    ///
+    pub fn is_guarded(&self) -> bool {
+        let all_variables = self.variables();
+        self.literals().any(|lit| {
+            let lit_variables: HashSet<&VariableRef> =
+                HashSet::from_iter(lit.variables().into_iter());
+            lit_variables == all_variables
+        })
+    }
+
+    ///
+    /// Frontier-Guarded $\small\text{Datalog}$ is defined where for every rule, all the variables
+    /// that are shared between the rule body and the rule head (called the frontier variables) must
+    /// all occur together in a guard atom.
+    ///
+    pub fn is_frontier_guarded(&self) -> bool {
+        let frontier_variables: HashSet<&VariableRef> = self
+            .head_variables()
+            .intersection(&self.variables())
+            .copied()
+            .collect();
+        self.literals().any(|lit| {
+            let lit_variables: HashSet<&VariableRef> =
+                HashSet::from_iter(lit.variables().into_iter());
+            lit_variables == frontier_variables
+        })
+    }
+
+    // --------------------------------------------------------------------------------------------
+
+    ///
+    /// # Rules Safety
     ///
     /// A Datalog program is well‐formed if all of its rules are well‐formed.
     ///
@@ -754,7 +781,14 @@ impl Rule {
     ///    2. all variables that appear in a negative literal in the body of a clause also appears
     ///       in some positive literal in the body of the clause.
     ///
-    pub fn well_formed_check(&self, features: &FeatureSet) -> Result<()> {
+    /// # Example unsafe rules
+    ///
+    /// ```datalog
+    /// [1]  invalid(X, Y) :- movie(X, Y, "1940"), Y > "1910".
+    /// [2]  invalid(X)    :- movie(X, Y, "1940"), NOT cast(U, X).
+    /// ```
+    ///
+    pub fn safety_check(&self, features: &FeatureSet) -> Result<()> {
         let (min, max) = if features.supports(&FEATURE_DISJUNCTION) {
             (1, usize::MAX)
         } else if features.supports(&FEATURE_CONSTRAINTS) {
@@ -792,7 +826,7 @@ impl Rule {
                 .collect();
             if !missing.is_empty() {
                 return Err(head_variables_missing_in_body(
-                    atom.label_ref().clone(),
+                    atom.label_ref(),
                     missing
                         .iter()
                         .map(|t| t.to_string())
@@ -804,11 +838,17 @@ impl Rule {
             let missing: Vec<&Term> = self
                 .negative_terms()
                 .into_iter()
-                .filter(|term| !body_positive_terms.contains(term))
+                .filter(|term| {
+                    if term.is_variable() {
+                        !body_positive_terms.contains(term)
+                    } else {
+                        false
+                    }
+                })
                 .collect();
             if !missing.is_empty() {
                 return Err(negative_variables_not_also_positive(
-                    atom.label_ref().clone(),
+                    atom.label_ref(),
                     missing
                         .iter()
                         .map(|t| t.to_string())
@@ -845,13 +885,9 @@ impl Display for Atom {
 }
 
 impl Labeled for Atom {
-    fn label(&self) -> &Predicate {
-        &self.label
-    }
+    get!(label -> Predicate);
 
-    fn label_ref(&self) -> PredicateRef {
-        self.label.clone()
-    }
+    get_cloned!(label_ref, label -> PredicateRef);
 }
 
 impl MaybeGround for Atom {
@@ -861,13 +897,9 @@ impl MaybeGround for Atom {
 }
 
 impl Collection<Term> for Atom {
-    fn is_empty(&self) -> bool {
-        self.terms.is_empty()
-    }
+    delegate!(is_empty, terms -> bool);
 
-    fn len(&self) -> usize {
-        self.terms.len()
-    }
+    delegate!(len, terms -> usize);
 
     fn iter(&self) -> Box<dyn Iterator<Item = &'_ Term> + '_> {
         Box::new(self.terms.iter())
@@ -903,7 +935,7 @@ impl Atom {
                         .unwrap()
                 {
                     return Err(fact_does_not_correspond_to_schema(
-                        relation.label_ref().clone(),
+                        relation.label_ref(),
                         terms
                             .iter()
                             .map(|t| t.to_string())
@@ -944,7 +976,7 @@ impl Atom {
 
     // --------------------------------------------------------------------------------------------
 
-    pub fn variables(&self) -> impl Iterator<Item = &Variable> {
+    pub fn variables(&self) -> impl Iterator<Item = &VariableRef> {
         self.terms.iter().filter_map(|t| {
             if let Term::Variable(v) = t {
                 Some(v)
@@ -954,9 +986,36 @@ impl Atom {
         })
     }
 
+    pub fn variable_index(&self, variable: &VariableRef) -> Option<usize> {
+        self.terms
+            .iter()
+            .enumerate()
+            .filter_map(|(i, term)| term.as_variable().map(|v| (i, v)))
+            .find(|(_, var)| var == &variable)
+            .map(|(i, _)| i)
+    }
+
+    pub fn constants(&self) -> impl Iterator<Item = &Constant> {
+        self.terms.iter().filter_map(|t| {
+            if let Term::Constant(v) = t {
+                Some(v)
+            } else {
+                None
+            }
+        })
+    }
+
     pub fn source_location(&self) -> Option<&SourceLocation> {
         self.src_loc.as_ref()
     }
+
+    pub fn is_existential(&self) -> bool {
+        self.constants().count() == self.len()
+    }
+
+    pub fn is_universal(&self) -> bool {
+        self.variables().count() == self.len()
+    }
 }
 
 // ------------------------------------------------------------------------------------------------
@@ -967,7 +1026,7 @@ impl Display for Literal {
             f,
             "{}{}",
             if self.negative {
-                NOT_UNICODE_SYMBOL
+                NEGATION_UNICODE_SYMBOL
             } else {
                 EMPTY_STR
             },
@@ -978,13 +1037,13 @@ impl Display for Literal {
 
 impl From<Atom> for Literal {
     fn from(v: Atom) -> Self {
-        Self::atom(v)
+        Self::relational(v)
     }
 }
 
 impl From<Comparison> for Literal {
     fn from(v: Comparison) -> Self {
-        Self::comparison(v)
+        Self::arithmetic(v)
     }
 }
 
@@ -1011,42 +1070,42 @@ impl Literal {
         Self { negative, inner }
     }
 
-    pub fn atom(atom: Atom) -> Self {
+    pub fn relational(atom: Atom) -> Self {
         Self::new(false, atom.into())
     }
 
-    pub fn negative_atom(atom: Atom) -> Self {
+    pub fn negative_relational(atom: Atom) -> Self {
         Self::new(true, atom.into())
     }
 
-    pub fn comparison(comparison: Comparison) -> Self {
+    pub fn arithmetic(comparison: Comparison) -> Self {
         Self::new(false, comparison.into())
     }
 
-    pub fn negative_comparison(comparison: Comparison) -> Self {
+    pub fn negative_arithmetic(comparison: Comparison) -> Self {
         Self::new(true, comparison.into())
     }
 
     // --------------------------------------------------------------------------------------------
 
-    delegate!(is_atom, inner -> bool);
+    delegate!(pub is_relational, inner -> bool);
 
-    delegate!(as_atom, inner -> Option<&Atom>);
+    delegate!(pub as_relational, inner -> Option<&Atom>);
 
-    delegate!(is_comparison, inner -> bool);
+    delegate!(pub is_arithmetic, inner -> bool);
 
-    delegate!(as_comparison, inner -> Option<&Comparison>);
+    delegate!(pub as_arithmetic, inner -> Option<&Comparison>);
 
     // --------------------------------------------------------------------------------------------
 
     pub fn terms(&self) -> Vec<&Term> {
         match &self.inner {
-            LiteralInner::Atom(v) => v.iter().collect(),
-            LiteralInner::Comparison(v) => v.terms(),
+            LiteralInner::Relational(v) => v.iter().collect(),
+            LiteralInner::Arithmetic(v) => v.terms(),
         }
     }
 
-    pub fn variables(&self) -> Vec<&Variable> {
+    pub fn variables(&self) -> Vec<&VariableRef> {
         self.terms()
             .iter()
             .filter_map(|t| {
@@ -1058,6 +1117,8 @@ impl Literal {
             })
             .collect()
     }
+
+    into_inner_fn!(LiteralInner, inner);
 }
 
 // ------------------------------------------------------------------------------------------------
@@ -1068,38 +1129,30 @@ impl Display for LiteralInner {
             f,
             "{}",
             match self {
-                Self::Atom(v) => v.to_string(),
-                Self::Comparison(v) => v.to_string(),
+                Self::Relational(v) => v.to_string(),
+                Self::Arithmetic(v) => v.to_string(),
             }
         )
     }
 }
 
-impl From<Atom> for LiteralInner {
-    fn from(v: Atom) -> Self {
-        Self::Atom(v)
-    }
-}
+impl_enum_from!(LiteralInner, Atom, Relational);
 
-impl From<Comparison> for LiteralInner {
-    fn from(v: Comparison) -> Self {
-        Self::Comparison(v)
-    }
-}
+impl_enum_from!(LiteralInner, Comparison, Arithmetic);
 
 impl MaybeGround for LiteralInner {
     fn is_ground(&self) -> bool {
         match self {
-            LiteralInner::Atom(a) => a.is_ground(),
-            LiteralInner::Comparison(c) => c.is_ground(),
+            LiteralInner::Relational(a) => a.is_ground(),
+            LiteralInner::Arithmetic(c) => c.is_ground(),
         }
     }
 }
 
 impl LiteralInner {
-    self_is_as!(atom, Atom);
+    self_is_as!(relational, Relational, Atom);
 
-    self_is_as!(comparison, Comparison);
+    self_is_as!(arithmetic, Arithmetic, Comparison);
 }
 
 // ------------------------------------------------------------------------------------------------
@@ -1116,7 +1169,21 @@ impl MaybeGround for Comparison {
 }
 
 impl Comparison {
-    pub fn new<L: Into<Term>, R: Into<Term>>(lhs: L, operator: ComparisonOperator, rhs: R) -> Self {
+    pub fn new<L: Into<Term>, R: Into<Term>>(
+        lhs: L,
+        operator: ComparisonOperator,
+        rhs: R,
+    ) -> Result<Self> {
+        let new_self = Self::new_unchecked(lhs, operator, rhs);
+        new_self.sanity_check()?;
+        Ok(new_self)
+    }
+
+    pub fn new_unchecked<L: Into<Term>, R: Into<Term>>(
+        lhs: L,
+        operator: ComparisonOperator,
+        rhs: R,
+    ) -> Self {
         Self {
             lhs: lhs.into(),
             operator,
@@ -1124,37 +1191,88 @@ impl Comparison {
         }
     }
 
-    pub fn eq<L: Into<Term>, R: Into<Term>>(left: L, right: R) -> Self {
+    pub fn sanity_check(&self) -> Result<()> {
+        match (
+            &self.lhs,
+            self.operator == ComparisonOperator::Equal,
+            &self.rhs,
+        ) {
+            (Term::Anonymous, _, _) | (_, _, Term::Anonymous) => {
+                Err(anonymous_variable_not_allowed())
+            }
+            (Term::Constant(lhs), true, Term::Constant(rhs)) => {
+                if lhs.kind() != rhs.kind() {
+                    Err(incompatible_types(
+                        lhs.kind().to_string(),
+                        rhs.kind().to_string(),
+                    ))
+                } else if lhs == rhs {
+                    Err(comparison_is_always_true(self.to_string()))
+                } else {
+                    Err(comparison_is_always_false(self.to_string()))
+                }
+            }
+            (Term::Constant(lhs), false, Term::Constant(rhs)) => {
+                if lhs.kind() != rhs.kind() {
+                    Err(incompatible_types(
+                        lhs.kind().to_string(),
+                        rhs.kind().to_string(),
+                    ))
+                } else if lhs == rhs {
+                    Err(comparison_is_always_false(self.to_string()))
+                } else {
+                    Err(comparison_is_always_true(self.to_string()))
+                }
+            }
+            (Term::Variable(lhs), true, Term::Variable(rhs)) => {
+                if lhs == rhs {
+                    Err(comparison_is_always_true(self.to_string()))
+                } else {
+                    Ok(())
+                }
+            }
+            (Term::Variable(lhs), false, Term::Variable(rhs)) => {
+                if lhs == rhs {
+                    Err(comparison_is_always_false(self.to_string()))
+                } else {
+                    Ok(())
+                }
+            }
+            _ => Ok(()),
+        }
+    }
+
+    pub fn eq<L: Into<Term>, R: Into<Term>>(left: L, right: R) -> Result<Self> {
         Self::new(left, ComparisonOperator::Equal, right)
     }
 
-    pub fn ne<L: Into<Term>, R: Into<Term>>(left: L, right: R) -> Self {
+    pub fn ne<L: Into<Term>, R: Into<Term>>(left: L, right: R) -> Result<Self> {
         Self::new(left, ComparisonOperator::NotEqual, right)
     }
 
-    pub fn lt<L: Into<Term>, R: Into<Term>>(left: L, right: R) -> Self {
+    pub fn lt<L: Into<Term>, R: Into<Term>>(left: L, right: R) -> Result<Self> {
         Self::new(left, ComparisonOperator::LessThan, right)
     }
 
-    pub fn lte<L: Into<Term>, R: Into<Term>>(left: L, right: R) -> Self {
+    pub fn lte<L: Into<Term>, R: Into<Term>>(left: L, right: R) -> Result<Self> {
         Self::new(left, ComparisonOperator::LessThanOrEqual, right)
     }
 
-    pub fn gt<L: Into<Term>, R: Into<Term>>(left: L, right: R) -> Self {
+    pub fn gt<L: Into<Term>, R: Into<Term>>(left: L, right: R) -> Result<Self> {
         Self::new(left, ComparisonOperator::GreaterThan, right)
     }
 
-    pub fn gte<L: Into<Term>, R: Into<Term>>(left: L, right: R) -> Self {
+    pub fn gte<L: Into<Term>, R: Into<Term>>(left: L, right: R) -> Result<Self> {
         Self::new(left, ComparisonOperator::GreaterThanOrEqual, right)
     }
 
     // --------------------------------------------------------------------------------------------
 
-    get!(lhs -> Term);
+    get!(pub lhs -> Term);
 
-    get!(rhs -> Term);
+    get!(pub rhs -> Term);
 
-    get!(operator -> ComparisonOperator);
+    get!(pub operator -> ComparisonOperator);
 
     // --------------------------------------------------------------------------------------------
 
@@ -1171,12 +1289,13 @@ impl Display for ComparisonOperator {
             f,
             "{}",
             match self {
-                Self::Equal => OPERATOR_ASCII_EQUAL,
-                Self::NotEqual => OPERATOR_ASCII_NOT_EQUAL,
-                Self::LessThan => OPERATOR_ASCII_LESS_THAN,
-                Self::LessThanOrEqual => OPERATOR_ASCII_LESS_THAN_OR_EQUAL,
-                Self::GreaterThan => OPERATOR_ASCII_GREATER_THAN,
-                Self::GreaterThanOrEqual => OPERATOR_ASCII_GREATER_THAN_OR_EQUAL,
+                Self::Equal => OPERATOR_EQUAL_ASCII,
+                Self::NotEqual => OPERATOR_NOT_EQUAL_ASCII,
+                Self::LessThan => OPERATOR_LESS_THAN_ASCII,
+                Self::LessThanOrEqual => OPERATOR_LESS_THAN_OR_EQUAL_ASCII,
+                Self::GreaterThan => OPERATOR_GREATER_THAN_ASCII,
+                Self::GreaterThanOrEqual => OPERATOR_GREATER_THAN_OR_EQUAL_ASCII,
+                Self::StringMatch => OPERATOR_STRING_MATCH_ASCII,
             }
         )
     }
@@ -1187,23 +1306,40 @@ impl FromStr for ComparisonOperator {
 
     fn from_str(s: &str) -> std::result::Result<Self, Self::Err> {
         match s {
-            OPERATOR_ASCII_EQUAL => Ok(Self::Equal),
-            OPERATOR_ASCII_NOT_EQUAL
-            | OPERATOR_ASCII_NOT_EQUAL_ALT
-            | OPERATOR_UNICODE_NOT_EQUAL => Ok(Self::NotEqual),
-            OPERATOR_ASCII_LESS_THAN => Ok(Self::LessThan),
-            OPERATOR_ASCII_LESS_THAN_OR_EQUAL | OPERATOR_UNICODE_LESS_THAN_OR_EQUAL => {
+            OPERATOR_EQUAL_ASCII => Ok(Self::Equal),
+            OPERATOR_NOT_EQUAL_ASCII
+            | OPERATOR_NOT_EQUAL_ASCII_ALT
+            | OPERATOR_NOT_EQUAL_UNICODE => Ok(Self::NotEqual),
+            OPERATOR_LESS_THAN_ASCII => Ok(Self::LessThan),
+            OPERATOR_LESS_THAN_OR_EQUAL_ASCII | OPERATOR_LESS_THAN_OR_EQUAL_UNICODE => {
                 Ok(Self::LessThanOrEqual)
             }
-            OPERATOR_ASCII_GREATER_THAN => Ok(Self::GreaterThan),
-            OPERATOR_ASCII_GREATER_THAN_OR_EQUAL | OPERATOR_UNICODE_GREATER_THAN_OR_EQUAL => {
+            OPERATOR_GREATER_THAN_ASCII => Ok(Self::GreaterThan),
+            OPERATOR_GREATER_THAN_OR_EQUAL_ASCII | OPERATOR_GREATER_THAN_OR_EQUAL_UNICODE => {
                 Ok(Self::GreaterThanOrEqual)
             }
+            OPERATOR_STRING_MATCH_ASCII
+            | OPERATOR_STRING_MATCH_UNICODE
+            | OPERATOR_STRING_MATCH_ASCII_WORD => Ok(Self::StringMatch),
             _ => Err(invalid_value(TYPE_NAME_COMPARISON_OPERATOR, s)),
         }
     }
 }
 
+impl ComparisonOperator {
+    pub fn inverse(&self) -> Self {
+        match self {
+            Self::Equal => Self::Equal,
+            Self::NotEqual => Self::NotEqual,
+            Self::LessThan => Self::GreaterThanOrEqual,
+            Self::LessThanOrEqual => Self::GreaterThan,
+            Self::GreaterThan => Self::LessThanOrEqual,
+            Self::GreaterThanOrEqual => Self::LessThan,
+            Self::StringMatch => Self::StringMatch,
+        }
+    }
+}
+
 // ------------------------------------------------------------------------------------------------
 
 impl Display for Term {
@@ -1220,17 +1356,9 @@ impl Display for Term {
     }
 }
 
-impl From<Variable> for Term {
-    fn from(v: Variable) -> Self {
-        Self::Variable(v)
-    }
-}
+impl_enum_from!(Term, VariableRef, Variable);
 
-impl From<Constant> for Term {
-    fn from(v: Constant) -> Self {
-        Self::Constant(v)
-    }
-}
+impl_enum_from!(Term, Constant, Constant);
 
 impl From<&str> for Term {
     fn from(s: &str) -> Self {
@@ -1267,7 +1395,7 @@ impl MaybeAnonymous for Term {
 }
 
 impl Term {
-    self_is_as!(variable, Variable);
+    self_is_as!(variable, Variable, VariableRef);
 
     self_is_as!(constant, Constant);
 }
@@ -1298,12 +1426,6 @@ impl AsRef<str> for Variable {
     }
 }
 
-impl From<Variable> for String {
-    fn from(v: Variable) -> Self {
-        v.0
-    }
-}
-
 impl AttributeName for Variable {
     fn is_valid(s: &str) -> bool {
         let mut chars = s.chars();
@@ -1312,24 +1434,43 @@ impl AttributeName for Variable {
             && chars.next().map(|c| c.is_uppercase()).unwrap()
             && chars.all(|c| c.is_alphanumeric() || c == CHAR_UNDERSCORE)
     }
+
+    fn type_name() -> &'static str {
+        TYPE_NAME_VARIABLE
+    }
 }
 
 impl Variable {
-    pub(crate) fn from_str_unchecked(s: &str) -> Self {
-        Self(s.to_owned())
-    }
+    into_inner_fn!(pub String);
 }
 
 // ------------------------------------------------------------------------------------------------
 // Private Functions
 // ------------------------------------------------------------------------------------------------
 
+//
+// TODO: make this better, it's horrible.
+//
+fn dedup_in_place<T: Clone + Eq + Hash, V: Into<Vec<T>>>(v: V) -> Vec<T> {
+    let v = v.into();
+    if v.len() <= 1 {
+        v
+    } else {
+        let mut vs: HashSet<T> = Default::default();
+        let mut v2: Vec<T> = Default::default();
+        for a in v.into_iter() {
+            if vs.insert(a.clone()) {
+                v2.push(a)
+            }
+        }
+        v2
+    }
+}
+
 // ------------------------------------------------------------------------------------------------
 // Modules
 // ------------------------------------------------------------------------------------------------
 
-mod eval;
-pub use eval::{naive::NaiveEvaluator, Evaluator, NoopEvaluator};
+pub mod eval;
 
-mod query;
-pub use query::{Query, Row, View};
+pub mod query;
