@@ -18,7 +18,7 @@ When referring to the specifics of the language we will use the common format $\
 superscripts that identify specific language extensions; for example, $\small\text{Datalog}^{\lnot}$ is
 the language extended with negation of literals, $\small\text{Datalog}^{\Gamma}$ is the language
 extended with type checking on attributes, and $\small\text{Datalog}^{\lnot,\theta}$. is the language
-extended with negation of literals _and_ comparison expressions. The order of superscript symbols is
+extended with negation of literals _and_ arithmetic literals. The order of superscript symbols is
 irrelevant. Additionally, text in **bold** indicates a key concept in the language while text in
 _italics_ indicates a forward reference to such a concept.
 
@@ -82,6 +82,13 @@ A pure rule is termed a _simple conditional assertion_, a constraint rule is ter
 _unconditional assertion_, and a disjunctive rule is termed a _sequent_ (or simply _conditional
 assertion_).
 
+Alternatively, some literature defines a rule in the following form:
+
+$$\tag{ia}\small C_1 | C_2 | \ldots | C_n \leftarrow A_1, \ldots, A_m, \lnot B_1, \ldots, \lnot B_k$$
+
+Where this form shows the expanded head structure according to $\small\text{Datalog}^{\lor}$, and the
+set of negated literals according to $\small\text{Datalog}^{\lnot}$.
+
 ### Terms
 
 Terms, mentioned above, may be constant values or variables such that
@@ -209,12 +216,12 @@ rule's head to be considered true.
 1. and has the following properties:
    1. $\small relational\(l\)$ returns true if the literal argument is a relational literal.
    1. $\small arithmetic\(l\)$ returns true if the literal argument is a arithmetic literal.
-   1. $\small terms\(l\)$ returns either the set of terms in either the atom or comparison,
+   1. $\small terms\(l\)$ returns either the set of terms in a literal,
       $$\tag{xiii}\small
         terms(l) \coloneqq
         \begin{cases}
-          terms(l), &\text{if } atom(l) \\\\
-          \lbrace t_{lhs}, t_{rhs} \rbrace, &\text{if } comparison(l) \land \text{Datalog}^{\theta}
+          terms(l), &\text{if } relational(l) \\\\
+          \lbrace t_{lhs}, t_{rhs} \rbrace, &\text{if } arithmetic(l) \land \text{Datalog}^{\theta}
         \end{cases}$$
    1. $\small ground\(l\)$ returns true if its terms are all constants $\small \(\forall{t}\in terms\(l\); t \in \mathcal{C}\)$,
    1. $\small positive\(l\)$ in $\small\text{Datalog}^{\lnot}$ returns false if negated,
@@ -409,8 +416,8 @@ emperor(X, rome).
 
 ### Literals
 
-Any valid atom is also a valid _positive_ literal. The syntax below also allows for _negative_
-literals as well as comparison expressions as literals. Conjunction may be written with the Unicode
+Any valid atom is also a valid _positive relational_ literal. The syntax below also allows for _negative_
+literals as well as arithmetic expressions as literals. Conjunction may be written with the Unicode
 character `∧` (logical and `\u{2227}`).
 
 ```abnf
@@ -440,39 +447,53 @@ alive(X) :- NOT dead(X).
 alive(X) ⟵ ￢dead(X).
 ```
 
-### Comparisons
+### Arithmetic Literals
 
 The language feature `comparisons` corresponds to the language $\small\text{Datalog}^{\theta}$ and
 allows the use of arithmetic literals. Comparisons take place between two literals and are
-currently limited to a set of common operators.
+currently limited to a set of common operators. Note the addition of a string match operator, this
+is similar to the Perl `=~` and requires a string value/variable on the left and a string value or
+variable on the right that compiles to a valid Rust regular expression. Finally, the rule `named-term`
+disallows the use of anonymous variables in arithmetic literals.
 
 ```abnf
-comparison      = term operator term
-operator        = "=" / "!=" / "/=" / "≠" / "<" / "<=" / "≤" / ">" / ">=" / "≥"
+comparison      = named-term operator named-term
+named-term      = named-variable / constant
+operator        = "="
+                / "!=" /"/=" / "≠"
+                / "<"
+                / "<=" / "≤"
+                / ">"
+                / ">=" / "≥"
+                / "*=" / "≛" / "MATCHES"
 ```
 
-The Unicode characters `≠` (not equal to `\u{2260}`), `≤` (less-than or equal to `\u{2264}`), and
-`≥` (greater-than or equal to `\u{2265}`) may be substituted for the common comparison operators.
+The Unicode characters `≠` (not equal to `\u{2260}`), `≤` (less-than or equal to `\u{2264}`),
+`≥` (greater-than or equal to `\u{2265}`, and star equals `\u{e2899b}`) may be substituted for the
+common arithmetic and string operators.
 
-All comparison operations **must** be between terms of the some type, such that the property
+All arithmetic operations **must** be between terms of the some type, such that the property
 _compatible_ introduce above is defined as:
 
 $$\tag{xvi}\small compatible(\tau_{lhs}, \tau_{rhs}, \theta) \leftarrow \tau_{lhs} = \tau_{rhs}$$
 
 Additionally, some operators are not present for all types, as shown in the table below.
 
-| Type     | `=`, `≠`   | `<`, `≤`, `>`, `≥` |
-| -------- | ---------- | ------------------ |
-| String   | Yes        | Yes - lexical      |
-| Integer  | Yes        | Yes                |
-| Boolean  | Yes        | No                 |
+| Type     | `=`, `≠`   | `<`, `≤`, `>`, `≥` | `≛` |
+| -------- | ---------- | ------------------ | --- |
+| String   | Yes        | Yes - lexical      | Yes |
+| Integer  | Yes        | Yes                | No  |
+| Boolean  | Yes        | No                 | No  |
 
-The following is an example using comparison of some numeric attribute of the _car_ relation.
+The following is an example using arithmetic literals and the _car_ relation.
 
 ```datalog
 .feature(comparisons).
+.assert car(make: string, model: string, age: integer).
 
-antique(X) :- car(X, Y) AND Y > 50.
+antique(X, Y) :- car(X, Y, _) AND X *= "[dD]uesenberg".
+antique(X, Y) :- car(X, Y, _) AND Y = "model t".
+antique(X, Y) :- car(X, Y, Z) AND Z > 50.
 ```
 
 ### Queries
@@ -499,14 +520,14 @@ the _model_ attribute in the _car_ relation where the _make_ is "ford", and igno
 ```datalog
 .assert car(make: string, model: string, age: integer).
 
-car("ford", Model, _)?
+car("ford", X, _)?
 ```
 
 The results of this query would not include the age column:
 
 ```text
 +------------+
-| Model      |
+| model      |
 +============+
 | edge       |
 +------------+
@@ -525,7 +546,8 @@ The results of this query would not include the age column:
 
 ### Pragmas
 
-TBD.
+Pragmas are declarative statements meant for the parser and runtime tooling, they do not affect
+the meaning of the program itself.
 
 ```abnf
 pragma          = feature / assert / infer / input / output
@@ -625,7 +647,7 @@ Note in this example we allow the parser to identify the schema for the relation
 The following is the same example constructed via the ASDI library.
 
 ```rust
-use asdi::{PredicateSet, Program};
+use asdi::{NameReferenceSet, Program};
 use asdi::edb::{Attribute, Predicate};
 use asdi::idb::{Atom, Term, Variable};
 use std::str::FromStr;
@@ -641,7 +663,8 @@ let human = syllogism
     .unwrap();
 human.add_as_fact(["Socrates".into()]).unwrap();
 
-let var_x: Term = Variable::from_str("X").unwrap().into();
+let variables = syllogism.variables();
+let var_x: Term = variables.fetch("X").unwrap().into();
 
 syllogism
     .add_new_pure_rule(
@@ -726,16 +749,20 @@ The program will select all matching (in this case all) facts from the _mortal_
     dyn_drop,
 )]
 
-use crate::edb::{Attribute, Predicate, PredicateRef, Relation, Relations, Schema};
+use crate::edb::{Attribute, Predicate, PredicateRef, Relation, RelationSet, Schema};
 use crate::error::{
     extensional_predicate_in_rule_head, language_feature_unsupported, relation_does_not_exist,
     Result,
 };
-use crate::features::{FeatureSet, FEATURE_CONSTRAINTS, FEATURE_DISJUNCTION};
-use crate::idb::{Atom, Evaluator, Literal, Query, Rule, RuleForm, Rules, Term, Variable, View};
+use crate::features::{FeatureSet, FEATURE_CONSTRAINTS, FEATURE_DISJUNCTION, FEATURE_NEGATION};
+use crate::idb::eval::{Evaluator, PrecedenceGraph};
+use crate::idb::query::{Query, Queryable, View};
+use crate::idb::{Atom, Literal, Rule, RuleForm, RuleSet, Term, Variable, VariableRef};
 use std::cell::RefCell;
 use std::collections::{BTreeMap, HashSet};
 use std::fmt::{Debug, Display, Formatter};
+use std::path::PathBuf;
+use std::rc::Rc;
 use std::str::FromStr;
 
 // ------------------------------------------------------------------------------------------------
@@ -743,16 +770,42 @@ use std::str::FromStr;
 // ------------------------------------------------------------------------------------------------
 
 ///
-/// A program consists of a set of extensional [`Relations`], a set of intensional
-/// [`Relations`], a set of [`Rules`], and a set of [queries](Query).
+/// Core, readable, properties of a Datalog program.
+pub trait ProgramCore {
+    ///
+    /// Returns the set of features currently supported by this program.
+    ///
+    fn features(&self) -> &FeatureSet;
+
+    ///
+    /// Returns the current set of extensional relations.
+    ///
+    fn extensional(&self) -> &RelationSet;
+
+    ///
+    /// Returns the current set of intensional relations.
+    ///
+    fn intensional(&self) -> &RelationSet;
+
+    ///
+    /// Return an iterator over the rules in the intensional database.
+    ///
+    fn rules(&self) -> &RuleSet;
+}
+
+///
+/// A program consists of a set of extensional [`RelationSet`], a set of intensional
+/// [`RelationSet`], a set of [`RuleSet`], and a set of [queries](Query).
 ///  
 #[derive(Clone, Debug, Default, PartialEq, Eq)]
 pub struct Program {
+    from_file: Option<PathBuf>,
     features: FeatureSet,
-    predicates: PredicateSet,
-    asserted: Relations,
-    infer: Relations,
-    rules: Rules,
+    predicate_cache: NameReferenceSet<Predicate>,
+    variable_cache: NameReferenceSet<Variable>,
+    extensional: RelationSet,
+    intensional: RelationSet,
+    rules: RuleSet,
     queries: HashSet<Query>,
 }
 
@@ -771,8 +824,34 @@ pub struct Program {
 /// assert!(program.predicates().fetch("Not A Predicate").is_err());
 /// ```
 ///
-#[derive(Clone, Debug, Default, PartialEq, Eq)]
-pub struct PredicateSet(RefCell<BTreeMap<String, PredicateRef>>);
+#[derive(Clone, Debug, PartialEq, Eq)]
+pub struct NameReferenceSet<T>(RefCell<BTreeMap<String, AttributeNameRef<T>>>)
+where
+    T: AttributeName;
+
+///
+/// Attributes, the members of [Schema] are named using different types in [Relation]s and [View]s.
+/// This trait identifies the minimum set of implementations required for an attribute name.
+///
+pub trait AttributeName:
+    AsRef<str> + Clone + Debug + Display + FromStr + PartialEq + Eq + PartialOrd + Ord
+{
+    ///
+    /// Return `true` if the string `s` is a valid value for the implementing type, else `false`.
+    ///
+    fn is_valid(s: &str) -> bool;
+
+    ///
+    /// Return the type identifier for the implementation, used in format errors.
+    ///
+    fn type_name() -> &'static str;
+}
+
+///
+/// A reference type for attribute names.
+///
+#[allow(type_alias_bounds)]
+pub type AttributeNameRef<T: AttributeName> = Rc<T>;
 
 ///
 /// All collections of things in the library implement these basic methods.
@@ -814,17 +893,6 @@ pub trait MutableIndexedCollection<K, V>: IndexedCollection<K, V> {
     fn insert<I: Into<K>>(&mut self, index: I, value: V) -> Result<()>;
 }
 
-///
-/// Attributes, the members of [Schema] are named using different types in [Relation]s and [View]s.
-/// This trait identifies the minimum set of implementations required for an attribute name.
-///
-pub trait AttributeName: Clone + Debug + Display + PartialEq + Eq + PartialOrd + Ord {
-    ///
-    /// Return `true` if the string `s` is a valid value for the implementing type, else `false`.
-    ///
-    fn is_valid(s: &str) -> bool;
-}
-
 ///
 /// Implemented by types that have, for sure, a label. This type is mutually exclusive with
 /// [MaybeLabeled].
@@ -863,7 +931,7 @@ pub trait MaybeLabeled<T: AttributeName>: MaybeAnonymous {
     ///
     /// Returns this value's label, or `None` if anonymous.
     ///
-    fn label(&self) -> Option<&T>;
+    fn label(&self) -> Option<&AttributeNameRef<T>>;
 
     ///
     /// Returns `true` if this value has a label, else `false`.
@@ -902,37 +970,6 @@ pub trait MaybePositive {
     }
 }
 
-///
-/// Commonly used properties of programs and rules.
-///
-pub trait SyntaxFragments {
-    ///
-    /// Linear $\small\text{Datalog}$ is defined where rule bodies must consist of a single atom.
-    ///
-    /// $$\tag{i}\small linear\(r\) \coloneqq |body\(r\)| = 1 \land \forall l \in body\(r\)\medspace \(atom\(l\)\)$$
-    ///
-    fn is_linear(&self) -> bool;
-
-    ///
-    /// Guarded $\small\text{Datalog}$ is defined where for every rule, all the variables that occur
-    /// in the rule bodies must occur together in at least one atom, called a guard atom.
-    ///
-    fn is_guarded(&self) -> bool;
-
-    ///
-    /// Frontier-Guarded $\small\text{Datalog}$ is defined where for every rule, all the variables
-    /// that are shared between the rule body and the rule head (called the frontier variables) must
-    /// all occur together in a guard atom.
-    ///
-    fn is_frontier_guarded(&self) -> bool;
-
-    ///
-    /// Non-Recursive $\small\text{Datalog}$ is defined by disallowing recursion in the definition
-    /// of programs, and therefore rules.
-    ///
-    fn is_non_recursive(&self) -> bool;
-}
-
 // ------------------------------------------------------------------------------------------------
 // Implementations
 // ------------------------------------------------------------------------------------------------
@@ -944,21 +981,21 @@ impl Display for Program {
             writeln!(f)?;
         }
 
-        if !self.asserted.is_empty() {
-            for relation in self.asserted.iter() {
+        if !self.extensional.is_empty() {
+            for relation in self.extensional.iter() {
                 writeln!(f, "{}", relation.to_schema_decl(true, false))?;
             }
             writeln!(f)?;
         }
 
-        if !self.infer.is_empty() {
-            for relation in self.infer.iter() {
+        if !self.intensional.is_empty() {
+            for relation in self.intensional.iter() {
                 writeln!(f, "{}", relation.to_schema_decl(false, false))?;
             }
             writeln!(f)?;
         }
 
-        for db in [&self.infer, &self.asserted] {
+        for db in [&self.intensional, &self.extensional] {
             for relation in db.iter() {
                 if !relation.is_empty() {
                     for fact in relation.iter() {
@@ -985,31 +1022,33 @@ impl MaybePositive for Program {
     }
 }
 
-impl SyntaxFragments for Program {
-    fn is_linear(&self) -> bool {
-        self.rules().iter().all(|rule| rule.is_linear())
+impl ProgramCore for Program {
+    fn features(&self) -> &FeatureSet {
+        &self.features
     }
 
-    fn is_guarded(&self) -> bool {
-        self.rules().iter().all(|rule| rule.is_guarded())
+    fn extensional(&self) -> &RelationSet {
+        &self.extensional
     }
 
-    fn is_frontier_guarded(&self) -> bool {
-        self.rules().iter().all(|rule| rule.is_frontier_guarded())
+    fn intensional(&self) -> &RelationSet {
+        &self.intensional
     }
 
-    fn is_non_recursive(&self) -> bool {
-        self.rules().iter().all(|rule| rule.is_non_recursive())
+    fn rules(&self) -> &RuleSet {
+        &self.rules
     }
 }
 
 impl Program {
     pub fn new_with_features(features: FeatureSet) -> Self {
         Self {
+            from_file: None,
             features,
-            predicates: Default::default(),
-            asserted: Default::default(),
-            infer: Default::default(),
+            predicate_cache: Default::default(),
+            variable_cache: Default::default(),
+            extensional: Default::default(),
+            intensional: Default::default(),
             queries: Default::default(),
             rules: Default::default(),
         }
@@ -1020,36 +1059,40 @@ impl Program {
     // --------------------------------------------------------------------------------------------
 
     ///
-    /// Returns the set of features currently supported by this program.
+    /// If this program were read from a source file, this will return the path to the file that
+    /// was read, else `None`.
     ///
-    pub fn features(&self) -> &FeatureSet {
-        &self.features
+    pub fn source_file_path(&self) -> Option<&PathBuf> {
+        self.from_file.as_ref()
     }
 
+    pub(crate) fn set_source_file_path(&mut self, file_name: PathBuf) {
+        self.from_file = Some(file_name);
+    }
+
+    // --------------------------------------------------------------------------------------------
+
     pub(crate) fn features_mut(&mut self) -> &mut FeatureSet {
         &mut self.features
     }
 
     // --------------------------------------------------------------------------------------------
 
-    pub fn predicates(&self) -> &PredicateSet {
-        &self.predicates
+    pub fn predicates(&self) -> &NameReferenceSet<Predicate> {
+        &self.predicate_cache
+    }
+
+    pub fn variables(&self) -> &NameReferenceSet<Variable> {
+        &self.variable_cache
     }
 
     // --------------------------------------------------------------------------------------------
 
-    ///
-    /// Returns the current set of extensional relations.
-    ///
-    pub fn extensional(&self) -> &Relations {
-        &self.asserted
-    }
-
     ///
     /// Returns the current set of extensional relations in a mutable state.
     ///
-    pub fn extensional_mut(&mut self) -> &mut Relations {
-        &mut self.asserted
+    pub fn extensional_mut(&mut self) -> &mut RelationSet {
+        &mut self.extensional
     }
 
     ///
@@ -1060,7 +1103,7 @@ impl Program {
         label: PredicateRef,
         schema: V,
     ) -> Result<&mut Relation> {
-        let label = self.predicates.canonical(label);
+        let label = self.predicate_cache.canonical(label);
         self.extensional_mut()
             .add_new_relation(label, schema.into())
     }
@@ -1074,18 +1117,11 @@ impl Program {
 
     // --------------------------------------------------------------------------------------------
 
-    ///
-    /// Returns the current set of intensional relations.
-    ///
-    pub fn intensional(&self) -> &Relations {
-        &self.infer
-    }
-
     ///
     /// Returns the current set of intensional relations in a mutable state.
     ///
-    pub fn intensional_mut(&mut self) -> &mut Relations {
-        &mut self.infer
+    pub fn intensional_mut(&mut self) -> &mut RelationSet {
+        &mut self.intensional
     }
 
     ///
@@ -1096,7 +1132,7 @@ impl Program {
         label: PredicateRef,
         schema: V,
     ) -> Result<&mut Relation> {
-        let label = self.predicates.canonical(label);
+        let label = self.predicate_cache.canonical(label);
         self.intensional_mut()
             .add_new_relation(label, schema.into())
     }
@@ -1108,13 +1144,6 @@ impl Program {
         self.intensional_mut().add(relation)
     }
 
-    ///
-    /// Return an iterator over the rules in the intensional database.
-    ///
-    pub fn rules(&self) -> &Rules {
-        &self.rules
-    }
-
     // Note: there is no `rules_mut` as we cannot allow clients to add rules without going through
     // the program so that we can ensure schema updates.
 
@@ -1128,7 +1157,7 @@ impl Program {
         head_terms: H,
         body: B,
     ) -> Result<()> {
-        let head_label = self.predicates.canonical(head_label);
+        let head_label = self.predicate_cache.canonical(head_label);
         let rule = Rule::new_pure(Atom::new(head_label, head_terms), body);
         self.add_rule(rule)
     }
@@ -1158,7 +1187,7 @@ impl Program {
     /// Add the provided `rule` to the intensional database.
     ///
     pub fn add_rule(&mut self, rule: Rule) -> Result<()> {
-        rule.well_formed_check(self.features())?;
+        rule.safety_check(self.features())?;
 
         if rule.form() == RuleForm::Constraint && !self.features().supports(&FEATURE_CONSTRAINTS) {
             return Err(language_feature_unsupported(FEATURE_CONSTRAINTS));
@@ -1172,12 +1201,12 @@ impl Program {
             //
             // Update the database schema based on atoms found in the rule's head.
             //
-            if self.asserted.contains(atom.label()) {
+            if self.extensional.contains(atom.label()) {
                 return Err(extensional_predicate_in_rule_head(
-                    atom.label_ref().clone(),
+                    atom.label_ref(),
                     atom.source_location().cloned(),
                 ));
-            } else if !self.infer.contains(atom.label()) {
+            } else if !self.intensional.contains(atom.label()) {
                 let mut schema = Vec::with_capacity(atom.len());
                 for term in atom.iter() {
                     match term {
@@ -1195,10 +1224,10 @@ impl Program {
         Ok(())
     }
 
-    fn infer_attribute(&self, variable: &Variable, rule: &Rule) -> Attribute<Predicate> {
+    fn infer_attribute(&self, variable: &VariableRef, rule: &Rule) -> Attribute<Predicate> {
         let candidates: Vec<(&Predicate, usize)> = rule
             .literals()
-            .filter_map(Literal::as_atom)
+            .filter_map(Literal::as_relational)
             .filter_map(|a| {
                 a.iter()
                     .enumerate()
@@ -1215,6 +1244,57 @@ impl Program {
         Attribute::anonymous()
     }
 
+    ///
+    /// A recursive $\small\text{Datalog}$ program has rules that directly, or indirectly recurse.
+    ///
+    pub fn is_recursive(&self) -> bool {
+        PrecedenceGraph::from(self).is_recursive()
+    }
+
+    ///
+    /// In $\small\text{Datalog}^{\lnot}$ a program is _semi-positive_ **iff** the only literals that
+    /// are negated are EDB relations.
+    ///
+    pub fn is_semi_positive(&self) -> bool {
+        if self.features().supports(&FEATURE_NEGATION) {
+            PrecedenceGraph::from(self).is_semi_positive()
+        } else {
+            false
+        }
+    }
+
+    ///
+    /// Linear $\small\text{Datalog}$ is defined where rule bodies have **at most** one IDB relation.
+    ///
+    /// $$\tag{i}\small linear\(r\) \coloneqq \ldots$$
+    ///
+    pub fn is_linear(&self) -> bool {
+        self.rules().iter().all(|rule| {
+            rule.literals()
+                .filter_map(|literal| literal.as_relational())
+                .filter(|atom| self.intensional().contains(atom.label()))
+                .count()
+                <= 1
+        })
+    }
+
+    ///
+    /// Guarded $\small\text{Datalog}$ is defined where for every rule, all the variables that occur
+    /// in the rule bodies must occur together in at least one atom, called a guard atom.
+    ///
+    pub fn is_guarded(&self) -> bool {
+        self.rules().iter().all(|rule| rule.is_guarded())
+    }
+
+    ///
+    /// Frontier-Guarded $\small\text{Datalog}$ is defined where for every rule, all the variables
+    /// that are shared between the rule body and the rule head (called the frontier variables) must
+    /// all occur together in a guard atom.
+    ///
+    pub fn is_frontier_guarded(&self) -> bool {
+        self.rules().iter().all(|rule| rule.is_frontier_guarded())
+    }
+
     // --------------------------------------------------------------------------------------------
 
     ///
@@ -1232,7 +1312,7 @@ impl Program {
         label: PredicateRef,
         terms: T,
     ) -> Result<bool> {
-        let label = self.predicates.canonical(label);
+        let label = self.predicate_cache.canonical(label);
         let query = Query::new(label, terms);
         self.add_query(query)
     }
@@ -1255,6 +1335,9 @@ impl Program {
     /// Load any data required from external files. For each relation any attached a [FilePragma](io/struct.FilePragma.html)
     /// is used to load data into that relation.
     ///
+    /// All files will be loaded either relative to the current working directory, or if the program
+    /// was loaded from a source file relative to that same source file.
+    ///
     pub fn load_extensional_data(&mut self) -> Result<()> {
         for relation in self.extensional_mut().iter_mut() {
             relation.load_from_file()?;
@@ -1266,6 +1349,9 @@ impl Program {
     /// Store any data required to external files. For each relation any attached a [FilePragma](io/struct.FilePragma.html)
     /// is used to store the relation's facts into a file.
     ///
+    /// All files will be stored either relative to the current working directory, or if the program
+    /// was loaded from a source file relative to that same source file.
+    ///
     pub fn store_intensional_data(&mut self) -> Result<()> {
         for relation in self.intensional_mut().iter_mut() {
             relation.store_to_file()?;
@@ -1293,7 +1379,7 @@ impl Program {
         }
         let new_idb = evaluator.inference(self)?;
         println!("{:?}", new_idb);
-        self.intensional_mut().merge(new_idb)?;
+        self.intensional_mut().merge_from(new_idb)?;
         self.store_intensional_data()?;
         let results = self.eval_queries()?;
         for (query, view) in results {
@@ -1358,21 +1444,33 @@ impl Program {
         }
     }
 
-    fn inner_eval_query(&self, query: &Query, intensional: &Relations) -> Result<Option<View>> {
+    fn inner_eval_query(&self, query: &Query, intensional: &RelationSet) -> Result<Option<View>> {
         let label = query.as_ref().label_ref();
         if intensional.contains(&label) {
-            Ok(intensional.matches(query.as_ref()))
+            intensional.query(query)
         } else if self.extensional().contains(&label) {
-            Ok(self.extensional().matches(query.as_ref()))
+            self.extensional().query(query)
         } else {
-            Err(relation_does_not_exist(label.clone()))
+            Err(relation_does_not_exist(label))
         }
     }
 }
 
 // ------------------------------------------------------------------------------------------------
 
-impl PredicateSet {
+impl<T> Default for NameReferenceSet<T>
+where
+    T: AttributeName,
+{
+    fn default() -> Self {
+        Self(Default::default())
+    }
+}
+
+impl<T> NameReferenceSet<T>
+where
+    T: AttributeName,
+{
     ///
     /// Add the value `s` as a predicate in this set.
     ///
@@ -1408,12 +1506,14 @@ impl PredicateSet {
     ///
     /// This will fail if the value provided in `s` does not the check [Predicate::is_valid].
     ///
-    pub fn fetch<S: Into<String>>(&self, s: S) -> Result<PredicateRef> {
+    pub fn fetch<S: Into<String>>(&self, s: S) -> Result<AttributeNameRef<T>> {
         let s = s.into();
         let found = { self.0.borrow().get(&s).cloned() };
         match found {
             None => {
-                let predicate: PredicateRef = Predicate::from_str(&s)?.into();
+                let predicate: AttributeNameRef<T> = T::from_str(&s)
+                    .map_err(|_| error::invalid_value(T::type_name(), &s))?
+                    .into();
                 let _ = self.0.borrow_mut().insert(s, predicate.clone());
                 Ok(predicate)
             }
@@ -1429,8 +1529,9 @@ impl PredicateSet {
     /// is added to the set and returned.
     ///
     #[inline]
-    pub fn canonical(&self, p: PredicateRef) -> PredicateRef {
-        let found = { self.0.borrow().get(p.as_ref().as_ref()).cloned() };
+    pub fn canonical(&self, p: AttributeNameRef<T>) -> AttributeNameRef<T> {
+        let s: &str = p.as_ref().as_ref();
+        let found = { self.0.borrow().get(s).cloned() };
         match found {
             None => {
                 let _ = self.0.borrow_mut().insert(p.to_string(), p.clone());
