@@ -1,11 +1,13 @@
 use asdi::edb::{Constant, Predicate};
-use asdi::idb::{Atom, Term, Variable};
+use asdi::idb::query::Queryable;
+use asdi::idb::{Atom, Term};
 use asdi::parse::parse_str;
+use asdi::ProgramCore;
 use std::rc::Rc;
 use std::str::FromStr;
 
 #[test]
-fn test_matches() {
+fn test_simple_query() {
     let program = parse_str(
         r#"% --> .assert human(string).
             
@@ -24,14 +26,22 @@ mortal(X) <- human(X).
 
     let human = Rc::new(Predicate::from_str("human").unwrap());
 
-    let query_term = Atom::new(
+    let query_atom = Atom::new(
         human.clone(),
-        [Term::Variable(Variable::from_str("X").unwrap())],
+        [Term::Variable(program.variables().fetch("X").unwrap())],
     );
-    let results = program.extensional().matches(&query_term).unwrap();
+    let results = program
+        .extensional()
+        .query_atom(&query_atom)
+        .unwrap()
+        .unwrap();
     println!("{}", results);
 
-    let query_term = Atom::new(human, [Term::Constant(Constant::from("Socrates"))]);
-    let results = program.extensional().matches(&query_term).unwrap();
+    let query_atom = Atom::new(human, [Term::Constant(Constant::from("Socrates"))]);
+    let results = program
+        .extensional()
+        .query_atom(&query_atom)
+        .unwrap()
+        .unwrap();
     println!("{}", results);
 }
